"""Implementation of time module."""

from __future__ import annotations

from datetime import UTC, datetime, tzinfo
from typing import cast
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from ...cachedzoneinfo import CachedZoneInfo
from ...feature import Feature
from ...interfaces import Time as TimeInterface
from ...smart.smartmodule import allow_update_after
from ..smartcammodule import SmartCamModule


class Time(SmartCamModule, TimeInterface):
    """Implementation of device_local_time."""

    QUERY_GETTER_NAME = "getTimezone"
    QUERY_MODULE_NAME = "system"
    QUERY_SECTION_NAMES = "basic"

    _timezone: tzinfo = UTC
    _time: datetime

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                device=self._device,
                id="device_time",
                name="Device time",
                attribute_getter="time",
                container=self,
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        q = super().query()
        q["getClockStatus"] = {self.QUERY_MODULE_NAME: {"name": "clock_status"}}

        return q

    async def _post_update_hook(self) -> None:
        """Perform actions after a device update."""
        time_data = self.data["getClockStatus"]["system"]["clock_status"]
        timezone_data = self.data["getTimezone"]["system"]["basic"]
        zone_id = timezone_data["zone_id"]
        timestamp = time_data["seconds_from_1970"]
        try:
            # Zoneinfo will return a DST aware object
            tz: tzinfo = await CachedZoneInfo.get_cached_zone_info(zone_id)
        except ZoneInfoNotFoundError:
            # timezone string like: UTC+10:00
            timezone_str = timezone_data["timezone"]
            tz = cast(tzinfo, datetime.strptime(timezone_str[-6:], "%z").tzinfo)

        self._timezone = tz
        self._time = datetime.fromtimestamp(
            cast(float, timestamp),
            tz=tz,
        )

    @property
    def timezone(self) -> tzinfo:
        """Return current timezone."""
        return self._timezone

    @property
    def time(self) -> datetime:
        """Return device's current datetime."""
        return self._time

    @allow_update_after
    async def set_time(self, dt: datetime) -> dict:
        """Set device time."""
        if not dt.tzinfo:
            timestamp = dt.replace(tzinfo=self.timezone).timestamp()
        else:
            timestamp = dt.timestamp()

        lt = datetime.fromtimestamp(timestamp).isoformat().replace("T", " ")
        params = {"seconds_from_1970": int(timestamp), "local_time": lt}
        # Doesn't seem to update the time, perhaps because timing_mode is ntp
        res = await self.call("setTimezone", {"system": {"clock_status": params}})
        if (zinfo := dt.tzinfo) and isinstance(zinfo, ZoneInfo):
            tz_params = {"zone_id": zinfo.key}
            res = await self.call("setTimezone", {"system": {"basic": tz_params}})
        return res



"""Module for the IOT legacy IOT KASA protocol."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable
from pprint import pformat as pf
from typing import TYPE_CHECKING, Any

from ..deviceconfig import DeviceConfig
from ..exceptions import (
    AuthenticationError,
    KasaException,
    TimeoutError,
    _ConnectionError,
    _RetryableError,
)
from ..json import dumps as json_dumps
from ..transports import XorEncryption, XorTransport
from .protocol import BaseProtocol, mask_mac, redact_data

if TYPE_CHECKING:
    from ..transports import BaseTransport

_LOGGER = logging.getLogger(__name__)


def _mask_children(children: list[dict[str, Any]]) -> list[dict[str, Any]]:
    def mask_child(child: dict[str, Any], index: int) -> dict[str, Any]:
        result = {
            **child,
            "id": f"SCRUBBED_CHILD_DEVICE_ID_{index + 1}",
        }
        # Will leave empty aliases as blank
        if child.get("alias"):
            result["alias"] = f"#MASKED_NAME# {index + 1}"
        return result

    return [mask_child(child, index) for index, child in enumerate(children)]


REDACTORS: dict[str, Callable[[Any], Any] | None] = {
    "latitude": lambda x: 0,
    "longitude": lambda x: 0,
    "latitude_i": lambda x: 0,
    "longitude_i": lambda x: 0,
    "deviceId": lambda x: "REDACTED_" + x[9::],
    "children": _mask_children,
    "alias": lambda x: "#MASKED_NAME#" if x else "",
    "mac": mask_mac,
    "mic_mac": mask_mac,
    "ssid": lambda x: "#MASKED_SSID#" if x else "",
    "oemId": lambda x: "REDACTED_" + x[9::],
    "username": lambda _: "user@example.com",  # cnCloud
    "hwId": lambda x: "REDACTED_" + x[9::],
}


class IotProtocol(BaseProtocol):
    """Class for the legacy TPLink IOT KASA Protocol."""

    BACKOFF_SECONDS_AFTER_TIMEOUT = 1

    def __init__(
        self,
        *,
        transport: BaseTransport,
    ) -> None:
        """Create a protocol object."""
        super().__init__(transport=transport)

        self._query_lock = asyncio.Lock()
        self._redact_data = True

    async def query(self, request: str | dict, retry_count: int = 3) -> dict:
        """Query the device retrying for retry_count on failure."""
        if isinstance(request, dict):
            request = json_dumps(request)
            assert isinstance(request, str)  # noqa: S101

        async with self._query_lock:
            return await self._query(request, retry_count)

    async def _query(self, request: str, retry_count: int = 3) -> dict:
        for retry in range(retry_count + 1):
            try:
                return await self._execute_query(request, retry)
            except _ConnectionError as sdex:
                if retry >= retry_count:
                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                    raise sdex
                continue
            except AuthenticationError as auex:
                await self._transport.reset()
                _LOGGER.debug(
                    "Unable to authenticate with %s, not retrying", self._host
                )
                raise auex
            except _RetryableError as ex:
                if retry == 0:
                    _LOGGER.debug(
                        "Device %s got a retryable error, will retry %s times: %s",
                        self._host,
                        retry_count,
                        ex,
                    )
                await self._transport.reset()
                if retry >= retry_count:
                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                    raise ex
                continue
            except TimeoutError as ex:
                if retry == 0:
                    _LOGGER.debug(
                        "Device %s got a timeout error, will retry %s times: %s",
                        self._host,
                        retry_count,
                        ex,
                    )
                await self._transport.reset()
                if retry >= retry_count:
                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                    raise ex
                await asyncio.sleep(self.BACKOFF_SECONDS_AFTER_TIMEOUT)
                continue
            except KasaException as ex:
                await self._transport.reset()
                _LOGGER.debug(
                    "Unable to query the device: %s, not retrying: %s",
                    self._host,
                    ex,
                )
                raise ex

        # make mypy happy, this should never be reached..
        raise KasaException("Query reached somehow to unreachable")

    async def _execute_query(self, request: str, retry_count: int) -> dict:
        debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)

        if debug_enabled:
            _LOGGER.debug(
                "%s >> %s",
                self._host,
                request,
            )
        resp = await self._transport.send(request)

        if debug_enabled:
            data = redact_data(resp, REDACTORS) if self._redact_data else resp
            _LOGGER.debug(
                "%s << %s",
                self._host,
                pf(data),
            )
        return resp

    async def close(self) -> None:
        """Close the underlying transport."""
        await self._transport.close()


class _deprecated_TPLinkSmartHomeProtocol(IotProtocol):
    def __init__(
        self,
        host: str | None = None,
        *,
        port: int | None = None,
        timeout: int | None = None,
        transport: BaseTransport | None = None,
    ) -> None:
        """Create a protocol object."""
        if not host and not transport:
            raise KasaException("host or transport must be supplied")
        if not transport:
            config = DeviceConfig(
                host=host,  # type: ignore[arg-type]
                port_override=port,
                timeout=timeout or XorTransport.DEFAULT_TIMEOUT,
            )
            transport = XorTransport(config=config)
        super().__init__(transport=transport)

    @staticmethod
    def encrypt(request: str) -> bytes:
        """Encrypt a request for a TP-Link Smart Home Device.

        :param request: plaintext request data
        :return: ciphertext to be send over wire, in bytes
        """
        return XorEncryption.encrypt(request)

    @staticmethod
    def decrypt(ciphertext: bytes) -> str:
        """Decrypt a response of a TP-Link Smart Home Device.

        :param ciphertext: encrypted response data
        :return: plaintext response
        """
        return XorEncryption.decrypt(ciphertext)



from __future__ import annotations

from datetime import datetime

import pytest
from pytest_mock import MockerFixture

from kasa import Module
from kasa.smart import SmartDevice

from ...device_fixtures import get_parent_and_child_modules, parametrize

autooff = parametrize(
    "has autooff", component_filter="auto_off", protocol_filter={"SMART"}
)


@autooff
@pytest.mark.parametrize(
    ("feature", "prop_name", "type"),
    [
        ("auto_off_enabled", "enabled", bool),
        ("auto_off_minutes", "delay", int),
        ("auto_off_at", "auto_off_at", datetime | None),
    ],
)
async def test_autooff_features(
    dev: SmartDevice, feature: str, prop_name: str, type: type
):
    """Test that features are registered and work as expected."""
    autooff = next(get_parent_and_child_modules(dev, Module.AutoOff))
    assert autooff is not None

    prop = getattr(autooff, prop_name)
    assert isinstance(prop, type)

    feat = autooff._device.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)


@autooff
async def test_settings(dev: SmartDevice, mocker: MockerFixture):
    """Test autooff settings."""
    autooff = next(get_parent_and_child_modules(dev, Module.AutoOff))
    assert autooff

    enabled = autooff._device.features["auto_off_enabled"]
    assert autooff.enabled == enabled.value

    delay = autooff._device.features["auto_off_minutes"]
    assert autooff.delay == delay.value

    call = mocker.spy(autooff, "call")
    new_state = True

    await autooff.set_enabled(new_state)
    call.assert_called_with(
        "set_auto_off_config", {"enable": new_state, "delay_min": delay.value}
    )
    call.reset_mock()
    await dev.update()

    new_delay = 123

    await autooff.set_delay(new_delay)

    call.assert_called_with(
        "set_auto_off_config", {"enable": new_state, "delay_min": new_delay}
    )

    await dev.update()

    assert autooff.enabled == new_state
    assert autooff.delay == new_delay


@autooff
@pytest.mark.parametrize("is_timer_active", [True, False])
async def test_auto_off_at(
    dev: SmartDevice, mocker: MockerFixture, is_timer_active: bool
):
    """Test auto-off at sensor."""
    autooff = next(get_parent_and_child_modules(dev, Module.AutoOff))
    assert autooff

    autooff_at = autooff._device.features["auto_off_at"]

    mocker.patch.object(
        type(autooff),
        "is_timer_active",
        new_callable=mocker.PropertyMock,
        return_value=is_timer_active,
    )
    if is_timer_active:
        assert isinstance(autooff_at.value, datetime)
    else:
        assert autooff_at.value is None



"""Module for smart plugs (HS100, HS110, ..)."""

from __future__ import annotations

import logging
from typing import Any

from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..module import Module
from ..protocols import BaseProtocol
from .iotdevice import IotDevice, requires_update
from .modules import AmbientLight, Antitheft, Cloud, Led, Motion, Schedule, Time, Usage

_LOGGER = logging.getLogger(__name__)


class IotPlug(IotDevice):
    r"""Representation of a TP-Link Smart Plug.

    To initialize, you have to await :func:`update()` at least once.
    This will allow accessing the properties using the exposed properties.

    All changes to the device are done using awaitable methods,
    which will not change the cached values,
    but you must await :func:`update()` separately.

    Errors reported by the device are raised as :class:`KasaException`\s,
    and should be handled by the user of the library.

    Examples:
        >>> import asyncio
        >>> plug = IotPlug("127.0.0.1")
        >>> asyncio.run(plug.update())
        >>> plug.alias
        Bedroom Lamp Plug

        Setting the LED state:

        >>> asyncio.run(plug.set_led(True))
        >>> asyncio.run(plug.update())
        >>> plug.led
        True

    For more examples, see the :class:`Device` class.
    """

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.Plug

    async def _initialize_modules(self) -> None:
        """Initialize modules."""
        await super()._initialize_modules()
        self.add_module(Module.IotSchedule, Schedule(self, "schedule"))
        self.add_module(Module.IotUsage, Usage(self, "schedule"))
        self.add_module(Module.IotAntitheft, Antitheft(self, "anti_theft"))
        self.add_module(Module.Time, Time(self, "time"))
        self.add_module(Module.IotCloud, Cloud(self, "cnCloud"))
        self.add_module(Module.Led, Led(self, "system"))

    @property  # type: ignore
    @requires_update
    def is_on(self) -> bool:
        """Return whether device is on."""
        sys_info = self.sys_info
        return bool(sys_info["relay_state"])

    async def turn_on(self, **kwargs: Any) -> dict:
        """Turn the switch on."""
        return await self._query_helper("system", "set_relay_state", {"state": 1})

    async def turn_off(self, **kwargs: Any) -> dict:
        """Turn the switch off."""
        return await self._query_helper("system", "set_relay_state", {"state": 0})


class IotWallSwitch(IotPlug):
    """Representation of a TP-Link Smart Wall Switch."""

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.WallSwitch

    async def _initialize_modules(self) -> None:
        """Initialize modules."""
        await super()._initialize_modules()
        if (dev_name := self.sys_info["dev_name"]) and "PIR" in dev_name:
            self.add_module(Module.IotMotion, Motion(self, "smartlife.iot.PIR"))
            self.add_module(
                Module.IotAmbientLight, AmbientLight(self, "smartlife.iot.LAS")
            )



from kasa import Device, Module

from ...device_fixtures import device_iot


@device_iot
def test_cloud(dev: Device):
    cloud = dev.modules.get(Module.IotCloud)
    assert cloud
    info = cloud.info
    assert info
    assert isinstance(info.provisioned, int)
    assert cloud.is_connected == bool(info.cloud_connected)



import datetime
from unittest.mock import Mock

import pytest
from voluptuous import (
    All,
    Any,
    Coerce,
    Range,
    Schema,
)

from kasa import Device, DeviceType, EmeterStatus, Module
from kasa.interfaces.energy import Energy
from kasa.iot import IotStrip
from kasa.iot.modules.emeter import Emeter
from tests.conftest import has_emeter_iot, no_emeter_iot

CURRENT_CONSUMPTION_SCHEMA = Schema(
    Any(
        {
            "voltage_mv": Any(All(float, Range(min=0, max=300000)), int, None),
            "power_mw": Any(Coerce(float), None),
            "current_ma": Any(All(float), int, None),
            "energy_wh": Any(Coerce(float), None),
            "total_wh": Any(Coerce(float), None),
            "voltage": Any(All(float, Range(min=0, max=300)), None),
            "power": Any(Coerce(float), None),
            "current": Any(All(float), None),
            "total": Any(Coerce(float), None),
            "energy": Any(Coerce(float), None),
            "slot_id": Any(Coerce(int), None),
        },
        None,
    )
)


@no_emeter_iot
async def test_no_emeter(dev):
    assert not dev.has_emeter

    with pytest.raises(AttributeError):
        await dev.get_emeter_realtime()

    with pytest.raises(AttributeError):
        await dev.get_emeter_daily()
    with pytest.raises(AttributeError):
        await dev.get_emeter_monthly()
    with pytest.raises(AttributeError):
        await dev.erase_emeter_stats()


@has_emeter_iot
async def test_get_emeter_realtime(dev):
    emeter = dev.modules[Module.Energy]

    current_emeter = await emeter.get_status()
    # Check realtime query gets the same value as status property
    # iot _query_helper strips out the error code from module responses.
    # but it's not stripped out of the _modular_update queries.
    assert current_emeter == {k: v for k, v in emeter.status.items() if k != "err_code"}
    CURRENT_CONSUMPTION_SCHEMA(current_emeter)


@has_emeter_iot
@pytest.mark.requires_dummy
async def test_get_emeter_daily(dev):
    emeter = dev.modules[Module.Energy]

    assert await emeter.get_daily_stats(year=1900, month=1) == {}

    d = await emeter.get_daily_stats()
    assert len(d) > 0

    k, v = d.popitem()
    assert isinstance(k, int)
    assert isinstance(v, float)

    # Test kwh (energy, energy_wh)
    d = await emeter.get_daily_stats(kwh=False)
    k2, v2 = d.popitem()
    assert v * 1000 == v2


@has_emeter_iot
@pytest.mark.requires_dummy
async def test_get_emeter_monthly(dev):
    emeter = dev.modules[Module.Energy]

    assert await emeter.get_monthly_stats(year=1900) == {}

    d = await emeter.get_monthly_stats()
    assert len(d) > 0

    k, v = d.popitem()
    assert isinstance(k, int)
    assert isinstance(v, float)

    # Test kwh (energy, energy_wh)
    d = await emeter.get_monthly_stats(kwh=False)
    k2, v2 = d.popitem()
    assert v * 1000 == v2


@has_emeter_iot
async def test_emeter_status(dev):
    emeter = dev.modules[Module.Energy]

    d = await emeter.get_status()

    with pytest.raises(KeyError):
        assert d["foo"]

    assert d["power_mw"] == d["power"] * 1000
    # bulbs have only power according to tplink simulator.
    if (
        dev.device_type is not DeviceType.Bulb
        and dev.device_type is not DeviceType.LightStrip
    ):
        assert d["voltage_mv"] == d["voltage"] * 1000

        assert d["current_ma"] == d["current"] * 1000
        assert d["total_wh"] == d["total"] * 1000


@pytest.mark.skip("not clearing your stats..")
@has_emeter_iot
async def test_erase_emeter_stats(dev):
    emeter = dev.modules[Module.Energy]

    await emeter.erase_emeter()


@has_emeter_iot
async def test_current_consumption(dev):
    emeter = dev.modules[Module.Energy]
    x = emeter.current_consumption
    assert isinstance(x, float)
    assert x >= 0.0


async def test_emeterstatus_missing_current():
    """KL125 does not report 'current' for emeter."""
    regular = EmeterStatus(
        {"err_code": 0, "power_mw": 0, "total_wh": 13, "current_ma": 123}
    )
    assert regular["current"] == 0.123

    with pytest.raises(KeyError):
        regular["invalid_key"]

    missing_current = EmeterStatus({"err_code": 0, "power_mw": 0, "total_wh": 13})
    assert missing_current["current"] is None


async def test_emeter_daily():
    """Test fetching the emeter for today.

    This test uses inline data since the fixtures
    will not have data for the current day.
    """
    emeter_data = {
        "get_daystat": {
            "day_list": [{"day": 1, "energy_wh": 8, "month": 1, "year": 2023}],
            "err_code": 0,
        }
    }

    class MockEmeter(Emeter):
        @property
        def data(self):
            return emeter_data

    emeter = MockEmeter(Mock(), "emeter")
    now = datetime.datetime.now()
    emeter_data["get_daystat"]["day_list"].append(
        {"day": now.day, "energy_wh": 500, "month": now.month, "year": now.year}
    )
    assert emeter.consumption_today == 0.500


@has_emeter_iot
async def test_supported(dev: Device):
    energy_module = dev.modules.get(Module.Energy)
    assert energy_module

    info = (
        dev._last_update
        if not isinstance(dev, IotStrip)
        else dev.children[0].internal_state
    )
    emeter = info[energy_module._module]["get_realtime"]
    has_total = "total" in emeter or "total_wh" in emeter
    has_voltage_current = "voltage" in emeter or "voltage_mv" in emeter
    assert energy_module.supports(Energy.ModuleFeature.CONSUMPTION_TOTAL) is has_total
    assert (
        energy_module.supports(Energy.ModuleFeature.VOLTAGE_CURRENT)
        is has_voltage_current
    )
    assert energy_module.supports(Energy.ModuleFeature.PERIODIC_STATS) is True



import pytest

from kasa import Device, Module

from ...device_fixtures import parametrize

motion = parametrize(
    "is motion sensor", model_filter="T100", protocol_filter={"SMART.CHILD"}
)


@motion
@pytest.mark.parametrize(
    ("feature", "type"),
    [
        ("motion_detected", bool),
    ],
)
async def test_motion_features(dev: Device, feature, type):
    """Test that features are registered and work as expected."""
    motion = dev.modules.get(Module.MotionSensor)
    assert motion is not None

    prop = getattr(motion, feature)
    assert isinstance(prop, type)

    feat = dev.features[feature]
    assert feat.value == prop
    assert isinstance(feat.value, type)



"""Module for common iotdevice tests."""

import re
from datetime import datetime

import pytest
from voluptuous import (
    REMOVE_EXTRA,
    All,
    Any,
    Boolean,
    In,
    Invalid,
    Optional,
    Range,
    Schema,
)

from kasa import DeviceType, KasaException, Module
from kasa.iot import IotDevice
from kasa.iot.iotmodule import _merge_dict
from tests.conftest import get_device_for_fixture_protocol, handle_turn_on, turn_on
from tests.device_fixtures import device_iot, has_emeter_iot, no_emeter_iot
from tests.fakeprotocol_iot import FakeIotProtocol

TZ_SCHEMA = Schema(
    {"zone_str": str, "dst_offset": int, "index": All(int, Range(min=0)), "tz_str": str}
)


def check_mac(x):
    if re.match("[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", x.lower()):
        return x
    raise Invalid(x)


SYSINFO_SCHEMA = Schema(
    {
        "active_mode": In(["schedule", "none", "count_down"]),
        "alias": str,
        "dev_name": str,
        "deviceId": str,
        "feature": str,
        "fwId": str,
        "hwId": str,
        "hw_ver": str,
        "icon_hash": str,
        "led_off": Boolean,
        "latitude": Any(All(float, Range(min=-90, max=90)), 0, None),
        "latitude_i": Any(
            All(int, Range(min=-900000, max=900000)),
            All(float, Range(min=-900000, max=900000)),
            0,
            None,
        ),
        "longitude": Any(All(float, Range(min=-180, max=180)), 0, None),
        "longitude_i": Any(
            All(int, Range(min=-18000000, max=18000000)),
            All(float, Range(min=-18000000, max=18000000)),
            0,
            None,
        ),
        "mac": check_mac,
        "model": str,
        "oemId": str,
        "on_time": int,
        "relay_state": int,
        "rssi": Any(int, None),  # rssi can also be positive, see #54
        "sw_ver": str,
        "type": str,
        "mic_type": str,
        "updating": Boolean,
        # these are available on hs220
        "brightness": int,
        "preferred_state": [
            {"brightness": All(int, Range(min=0, max=100)), "index": int}
        ],
        "next_action": {"type": int},
        "child_num": Optional(Any(None, int)),
        "children": Optional(list),
    },
    extra=REMOVE_EXTRA,
)


@device_iot
async def test_state_info(dev):
    assert isinstance(dev.state_information, dict)


@pytest.mark.requires_dummy
@device_iot
async def test_invalid_connection(mocker, dev):
    mocker.patch.object(FakeIotProtocol, "query", side_effect=KasaException)
    with pytest.raises(KasaException):
        await dev.update()


@has_emeter_iot
async def test_initial_update_emeter(dev, mocker):
    """Test that the initial update performs second query if emeter is available."""
    dev._last_update = {}
    dev._legacy_features = set()
    spy = mocker.spy(dev.protocol, "query")
    await dev.update()
    # Devices with small buffers may require 3 queries
    expected_queries = 2 if dev.max_device_response_size > 4096 else 3
    assert spy.call_count == expected_queries + len(dev.children)


@no_emeter_iot
async def test_initial_update_no_emeter(dev, mocker):
    """Test that the initial update performs second query if emeter is available."""
    dev._last_update = {}
    dev._legacy_features = set()
    spy = mocker.spy(dev.protocol, "query")
    await dev.update()
    # child calls will happen if a child has a module with a query (e.g. schedule)
    child_calls = 0
    for child in dev.children:
        for module in child.modules.values():
            if module.query():
                child_calls += 1
                break
    # 2 parent are necessary as some devices crash on unexpected modules
    # See #105, #120, #161
    assert spy.call_count == 2 + child_calls


@device_iot
async def test_query_helper(dev):
    with pytest.raises(KasaException):
        await dev._query_helper("test", "testcmd", {})
    # TODO check for unwrapping?


@device_iot
@turn_on
async def test_state(dev, turn_on):
    orig_state = dev.is_on
    await handle_turn_on(dev, turn_on)
    await dev.update()
    if orig_state:
        await dev.turn_off()
        await dev.update()
        assert not dev.is_on
        assert dev.is_off

        await dev.turn_on()
        await dev.update()
        assert dev.is_on
        assert not dev.is_off
    else:
        await dev.turn_on()
        await dev.update()
        assert dev.is_on
        assert not dev.is_off

        await dev.turn_off()
        await dev.update()
        assert not dev.is_on
        assert dev.is_off


@device_iot
@turn_on
async def test_on_since(dev, turn_on):
    await handle_turn_on(dev, turn_on)
    orig_state = dev.is_on
    if "on_time" not in dev.sys_info and dev.device_type is not DeviceType.Strip:
        assert dev.on_since is None
    elif orig_state:
        assert isinstance(dev.on_since, datetime)
    else:
        assert dev.on_since is None


@device_iot
async def test_time(dev):
    assert isinstance(dev.modules[Module.Time].time, datetime)


@device_iot
async def test_timezone(dev):
    TZ_SCHEMA(await dev.modules[Module.Time].get_timezone())


@device_iot
async def test_hw_info(dev):
    SYSINFO_SCHEMA(dev.hw_info)


@device_iot
async def test_location(dev):
    SYSINFO_SCHEMA(dev.location)


@device_iot
async def test_rssi(dev):
    SYSINFO_SCHEMA({"rssi": dev.rssi})  # wrapping for vol


@device_iot
async def test_mac(dev):
    SYSINFO_SCHEMA({"mac": dev.mac})  # wrapping for val


@device_iot
async def test_representation(dev):
    pattern = re.compile(r"<DeviceType\..+ at .+? - .*? \(.+?\)>")
    assert pattern.match(str(dev))


@device_iot
async def test_children(dev):
    """Make sure that children property is exposed by every device."""
    if dev.device_type is DeviceType.Strip:
        assert len(dev.children) > 0
    else:
        assert len(dev.children) == 0


@device_iot
async def test_modules_preserved(dev: IotDevice):
    """Make modules that are not being updated are preserved between updates."""
    dev._last_update["some_module_not_being_updated"] = "should_be_kept"
    await dev.update()
    assert dev._last_update["some_module_not_being_updated"] == "should_be_kept"


@device_iot
async def test_internal_state(dev):
    """Make sure the internal state returns the last update results."""
    assert dev.internal_state == dev._last_update


@device_iot
async def test_features(dev):
    """Make sure features is always accessible."""
    sysinfo = dev._last_update["system"]["get_sysinfo"]
    if "feature" in sysinfo:
        assert dev._legacy_features == set(sysinfo["feature"].split(":"))
    else:
        assert dev._legacy_features == set()


@device_iot
async def test_max_device_response_size(dev):
    """Make sure every device return has a set max response size."""
    assert dev.max_device_response_size > 0


@device_iot
async def test_estimated_response_sizes(dev):
    """Make sure every module has an estimated response size set."""
    for mod in dev.modules.values():
        assert mod.estimated_query_response_size > 0


@device_iot
async def test_modules_not_supported(dev: IotDevice):
    """Test that unsupported modules do not break the device."""
    for module in dev.modules.values():
        assert module.is_supported is not None
    await dev.update()
    for module in dev.modules.values():
        assert module.is_supported is not None


async def test_get_modules():
    """Test getting modules for child and parent modules."""
    dummy_device = await get_device_for_fixture_protocol(
        "HS100(US)_2.0_1.5.6.json", "IOT"
    )
    from kasa.iot.modules import Cloud

    # Modules on device
    module = dummy_device.modules.get("cloud")
    assert module
    assert module.device == dummy_device
    assert isinstance(module, Cloud)

    module = dummy_device.modules.get(Module.IotCloud)
    assert module
    assert module.device == dummy_device
    assert isinstance(module, Cloud)

    # Invalid modules
    module = dummy_device.modules.get("DummyModule")
    assert module is None

    module = dummy_device.modules.get(Module.Cloud)
    assert module is None


def test_merge_dict():
    """Test the recursive dict merge."""
    dest = {"a": 1, "b": {"c": 2, "d": 3}}
    source = {"b": {"c": 4, "e": 5}}
    assert _merge_dict(dest, source) == {"a": 1, "b": {"c": 4, "d": 3, "e": 5}}

    dest = {"smartlife.iot.common.emeter": {"get_realtime": None}}
    source = {
        "smartlife.iot.common.emeter": {"get_daystat": {"month": 8, "year": 2024}}
    }
    assert _merge_dict(dest, source) == {
        "smartlife.iot.common.emeter": {
            "get_realtime": None,
            "get_daystat": {"month": 8, "year": 2024},
        }
    }



"""Implementation of vacuum cleaning records."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime, timedelta, tzinfo
from typing import Annotated, cast

from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import ADD_DIALECT_SUPPORT
from mashumaro.dialect import Dialect
from mashumaro.types import SerializationStrategy

from ...feature import Feature
from ...module import FeatureAttribute
from ..smartmodule import Module, SmartModule
from .clean import AreaUnit, Clean

_LOGGER = logging.getLogger(__name__)


@dataclass
class Record(DataClassDictMixin):
    """Historical cleanup result."""

    class Config:
        """Configuration class."""

        code_generation_options = [ADD_DIALECT_SUPPORT]

    #: Total time cleaned (in minutes)
    clean_time: timedelta = field(
        metadata=field_options(deserialize=lambda x: timedelta(minutes=x))
    )
    #: Total area cleaned
    clean_area: int
    dust_collection: bool
    timestamp: datetime

    info_num: int | None = None
    message: int | None = None
    map_id: int | None = None
    start_type: int | None = None
    task_type: int | None = None
    record_index: int | None = None

    #: Error code from cleaning
    error: int = field(default=0)


class _DateTimeSerializationStrategy(SerializationStrategy):
    def __init__(self, tz: tzinfo) -> None:
        self.tz = tz

    def deserialize(self, value: float) -> datetime:
        return datetime.fromtimestamp(value, self.tz)


def _get_tz_strategy(tz: tzinfo) -> type[Dialect]:
    """Return a timezone aware de-serialization strategy."""

    class TimezoneDialect(Dialect):
        serialization_strategy = {datetime: _DateTimeSerializationStrategy(tz)}

    return TimezoneDialect


@dataclass
class Records(DataClassDictMixin):
    """Response payload for getCleanRecords."""

    class Config:
        """Configuration class."""

        code_generation_options = [ADD_DIALECT_SUPPORT]

    total_time: timedelta = field(
        metadata=field_options(deserialize=lambda x: timedelta(minutes=x))
    )
    total_area: int
    total_count: int = field(metadata=field_options(alias="total_number"))

    records: list[Record] = field(metadata=field_options(alias="record_list"))
    last_clean: Record = field(metadata=field_options(alias="lastest_day_record"))

    @classmethod
    def __pre_deserialize__(cls, d: dict) -> dict:
        if ldr := d.get("lastest_day_record"):
            d["lastest_day_record"] = {
                "timestamp": ldr[0],
                "clean_time": ldr[1],
                "clean_area": ldr[2],
                "dust_collection": ldr[3],
            }
        return d


class CleanRecords(SmartModule):
    """Implementation of vacuum cleaning records."""

    REQUIRED_COMPONENT = "clean_percent"
    _parsed_data: Records

    async def _post_update_hook(self) -> None:
        """Cache parsed data after an update."""
        self._parsed_data = Records.from_dict(
            self.data, dialect=_get_tz_strategy(self._device.timezone)
        )

    def _initialize_features(self) -> None:
        """Initialize features."""
        for type_ in ["total", "last"]:
            self._add_feature(
                Feature(
                    self._device,
                    id=f"{type_}_clean_area",
                    name=f"{type_.capitalize()} area cleaned",
                    container=self,
                    attribute_getter=f"{type_}_clean_area",
                    unit_getter="area_unit",
                    category=Feature.Category.Debug,
                    type=Feature.Type.Sensor,
                )
            )
            self._add_feature(
                Feature(
                    self._device,
                    id=f"{type_}_clean_time",
                    name=f"{type_.capitalize()} time cleaned",
                    container=self,
                    attribute_getter=f"{type_}_clean_time",
                    category=Feature.Category.Debug,
                    type=Feature.Type.Sensor,
                )
            )
        self._add_feature(
            Feature(
                self._device,
                id="total_clean_count",
                name="Total clean count",
                container=self,
                attribute_getter="total_clean_count",
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="last_clean_timestamp",
                name="Last clean timestamp",
                container=self,
                attribute_getter="last_clean_timestamp",
                category=Feature.Category.Debug,
                type=Feature.Type.Sensor,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {
            "getCleanRecords": {},
        }

    @property
    def total_clean_area(self) -> Annotated[int, FeatureAttribute()]:
        """Return total cleaning area."""
        return self._parsed_data.total_area

    @property
    def total_clean_time(self) -> timedelta:
        """Return total cleaning time."""
        return self._parsed_data.total_time

    @property
    def total_clean_count(self) -> int:
        """Return total clean count."""
        return self._parsed_data.total_count

    @property
    def last_clean_area(self) -> Annotated[int, FeatureAttribute()]:
        """Return latest cleaning area."""
        return self._parsed_data.last_clean.clean_area

    @property
    def last_clean_time(self) -> timedelta:
        """Return total cleaning time."""
        return self._parsed_data.last_clean.clean_time

    @property
    def last_clean_timestamp(self) -> datetime:
        """Return latest cleaning timestamp."""
        return self._parsed_data.last_clean.timestamp

    @property
    def area_unit(self) -> AreaUnit:
        """Return area unit."""
        clean = cast(Clean, self._device.modules[Module.Clean])
        return clean.area_unit

    @property
    def parsed_data(self) -> Records:
        """Return parsed records data."""
        return self._parsed_data



"""Child device implementation."""

from __future__ import annotations

import logging
import time
from typing import Any

from ..device import DeviceInfo
from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..protocols.smartprotocol import SmartProtocol, _ChildProtocolWrapper
from .smartdevice import ComponentsRaw, SmartDevice
from .smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class SmartChildDevice(SmartDevice):
    """Presentation of a child device.

    This wraps the protocol communications and sets internal data for the child.
    """

    CHILD_DEVICE_TYPE_MAP = {
        "plug.powerstrip.sub-plug": DeviceType.Plug,
        "subg.plugswitch.switch": DeviceType.WallSwitch,
        "subg.trigger.contact-sensor": DeviceType.Sensor,
        "subg.trigger.temp-hmdt-sensor": DeviceType.Sensor,
        "subg.trigger.water-leak-sensor": DeviceType.Sensor,
        "subg.trigger.motion-sensor": DeviceType.Sensor,
        "kasa.switch.outlet.sub-fan": DeviceType.Fan,
        "kasa.switch.outlet.sub-dimmer": DeviceType.Dimmer,
        "subg.trv": DeviceType.Thermostat,
        "subg.trigger.button": DeviceType.Sensor,
    }

    def __init__(
        self,
        parent: SmartDevice,
        info: dict,
        component_info_raw: ComponentsRaw,
        *,
        config: DeviceConfig | None = None,
        protocol: SmartProtocol | None = None,
    ) -> None:
        self._id = info["device_id"]
        _protocol = protocol or _ChildProtocolWrapper(self._id, parent.protocol)
        super().__init__(parent.host, config=parent.config, protocol=_protocol)
        self._parent = parent
        self._update_internal_state(info)
        self._components_raw = component_info_raw
        self._components = self._parse_components(self._components_raw)

    @property
    def device_info(self) -> DeviceInfo:
        """Return device info.

        Child device does not have it info and components in _last_update so
        this overrides the base implementation to call _get_device_info with
        info and components combined as they would be in _last_update.
        """
        return self._get_device_info(
            {
                "get_device_info": self._info,
                "component_nego": self._components_raw,
            },
            None,
        )

    async def update(self, update_children: bool = True) -> None:
        """Update child module info.

        The parent updates our internal info so just update modules with
        their own queries.
        """
        await self._update(update_children)

    async def _update(self, update_children: bool = True) -> None:
        """Update child module info.

        Internal implementation to allow patching of public update in the cli
        or test framework.
        """
        now = time.monotonic()
        module_queries: list[SmartModule] = []
        req: dict[str, Any] = {}
        for module in self.modules.values():
            if (
                module.disabled is False
                and (mod_query := module.query())
                and module._should_update(now)
            ):
                module_queries.append(module)
                req.update(mod_query)
        if req:
            first_update = self._last_update != {}
            try:
                resp = await self.protocol.query(req)
            except Exception as ex:
                resp = await self._handle_modular_update_error(
                    ex, first_update, ", ".join(mod.name for mod in module_queries), req
                )
            self._last_update = resp

        for module in self.modules.values():
            await self._handle_module_post_update(
                module, now, had_query=module in module_queries
            )
        self._last_update_time = now

        # We can first initialize the features after the first update.
        # We make here an assumption that every device has at least a single feature.
        if not self._features:
            await self._initialize_features()

    @classmethod
    async def create(
        cls,
        parent: SmartDevice,
        child_info: dict,
        child_components_raw: ComponentsRaw,
        protocol: SmartProtocol | None = None,
        *,
        last_update: dict | None = None,
    ) -> SmartDevice:
        """Create a child device based on device info and component listing.

        If creating a smart child from a different protocol, i.e. a camera hub,
        protocol: SmartProtocol and last_update should be provided as per the
        FIRST_UPDATE_MODULES expected by the update cycle as these cannot be
        derived from the parent.
        """
        child: SmartChildDevice = cls(
            parent, child_info, child_components_raw, protocol=protocol
        )
        if last_update:
            child._last_update = last_update
        await child._initialize_modules()
        return child

    @property
    def device_type(self) -> DeviceType:
        """Return child device type."""
        if self._device_type is not DeviceType.Unknown:
            return self._device_type

        if self.sys_info and (category := self.sys_info.get("category")):
            dev_type = self.CHILD_DEVICE_TYPE_MAP.get(category)
            if dev_type is None:
                _LOGGER.warning(
                    "Unknown child device type %s for model %s, please open issue",
                    category,
                    self.model,
                )
                self._device_type = DeviceType.Unknown
            else:
                self._device_type = dev_type

        return self._device_type

    def __repr__(self) -> str:
        if not self._parent:
            return f"<{self.device_type}(child) without parent>"
        if not self._parent._last_update:
            return f"<{self.device_type}(child) of {self._parent}>"
        return f"<{self.device_type} {self.alias} ({self.model}) of {self._parent}>"



"""Interact with child devices.

>>> from kasa import Discover
>>>
>>> dev = await Discover.discover_single(
>>>     "127.0.0.1",
>>>     username="user@example.com",
>>>     password="great_password"
>>> )
>>> await dev.update()
>>> print(dev.alias)
Bedroom Power Strip

All methods act on the whole strip:

>>> for plug in dev.children:
>>>    print(f"{plug.alias}: {plug.is_on}")
Plug 1: True
Plug 2: False
Plug 3: False
>>> dev.is_on
True
>>> await dev.turn_off()
>>> await dev.update()

Accessing individual plugs can be done using the `children` property:

>>> len(dev.children)
3
>>> for plug in dev.children:
>>>    print(f"{plug.alias}: {plug.is_on}")
Plug 1: False
Plug 2: False
Plug 3: False
>>> await dev.children[1].turn_on()
>>> await dev.update()
>>> dev.is_on
True
"""

from ...device_type import DeviceType
from ..smartmodule import SmartModule


class ChildDevice(SmartModule):
    """Implementation for child devices."""

    REQUIRED_COMPONENT = "child_device"
    QUERY_GETTER_NAME = "get_child_device_list"

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        q = super().query()
        if self._device.device_type is DeviceType.Hub:
            q["get_child_device_component_list"] = None
        return q



#!/usr/bin/env python
"""
This code allow for the decryption of KlapV2 data from a pcap file.

It will output the decrypted data to a file.
This was designed and tested with a Tapo light strip setup using a cloud account.
"""

from __future__ import annotations

import asyncio
import codecs
import json
import re
from threading import Thread

import asyncclick as click
import pyshark
from cryptography.hazmat.primitives import padding

from kasa.credentials import DEFAULT_CREDENTIALS, Credentials, get_default_credentials
from kasa.deviceconfig import (
    DeviceConfig,
    DeviceConnectionParameters,
    DeviceEncryptionType,
    DeviceFamily,
)
from kasa.transports.klaptransport import KlapEncryptionSession, KlapTransportV2


def _get_seq_from_query(packet):
    """Return sequence number for the query."""
    query = packet.http.get("request_uri_query")
    if query is None:
        raise Exception("No request_uri_query found")
    # use regex to get: seq=(\d+)
    seq = re.search(r"seq=(\d+)", query)
    if seq is not None:
        return int(seq.group(1))
    raise Exception("Unable to find sequence number")


def _is_http_response_for_packet(response, packet):
    """Return True if the *response* contains a response for request in *packet*.

    Different tshark versions use different field for the information.
    """
    if not hasattr(response, "http"):
        return False
    if hasattr(response.http, "response_for_uri") and (
        response.http.response_for_uri == packet.http.request_full_uri
    ):
        return True
    # tshark 4.4.0
    return response.http.request_uri == packet.http.request_uri


class MyEncryptionSession(KlapEncryptionSession):
    """A custom KlapEncryptionSession class that allows for decryption."""

    def decrypt(self, msg):
        """Decrypt the data."""
        decryptor = self._cipher.decryptor()
        dp = decryptor.update(msg[32:]) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        plaintextbytes = unpadder.update(dp) + unpadder.finalize()

        return plaintextbytes.decode("utf-8", "bad_chars_replacement")


class Operator:
    """A class that handles the data decryption, and the encryption session updating."""

    def __init__(self, klap, creds):
        self._local_seed: bytes | None = None
        self._remote_seed: bytes | None = None
        self._session: MyEncryptionSession | None = None
        self._creds = creds
        self._klap: KlapTransportV2 = klap
        self._auth_hash = self._klap.generate_auth_hash(self._creds)
        self._local_seed_auth_hash = None
        self._remote_seed_auth_hash = None
        self._seq = 0

    def check_default_credentials(self):
        """Check whether default credentials were used.

        Devices sometimes randomly accept the hardcoded default credentials
        and the library handles that.
        """
        for value in DEFAULT_CREDENTIALS.values():
            default_credentials = get_default_credentials(value)
            default_auth_hash = self._klap.generate_auth_hash(default_credentials)
            default_credentials_seed_auth_hash = self._klap.handshake1_seed_auth_hash(
                self._local_seed,
                self._remote_seed,
                default_auth_hash,  # type: ignore
            )
            if self._remote_seed_auth_hash == default_credentials_seed_auth_hash:
                return default_auth_hash
        return None

    def update_encryption_session(self):
        """Update the encryption session used for decrypting data.

        It is called whenever the local_seed, remote_seed,
        or remote_auth_hash is updated.

        It checks if the seeds are set and, if they are, creates a new session.

        Raises:
            ValueError: If the auth hashes do not match.
        """
        if self._local_seed is None or self._remote_seed is None:
            self._session = None
        else:
            self._local_seed_auth_hash = self._klap.handshake1_seed_auth_hash(
                self._local_seed, self._remote_seed, self._auth_hash
            )
            auth_hash = None
            if self._remote_seed_auth_hash is not None:
                if self._local_seed_auth_hash == self._remote_seed_auth_hash:
                    auth_hash = self._auth_hash
                else:
                    auth_hash = self.check_default_credentials()
                if not auth_hash:
                    raise ValueError(
                        "Local and remote auth hashes do not match. "
                        "This could mean an incorrect username and/or password."
                    )
                self._session = MyEncryptionSession(
                    self._local_seed, self._remote_seed, auth_hash
                )
                self._session._seq = self._seq
                self._session._generate_cipher()

    @property
    def seq(self) -> int:
        """Get the sequence number."""
        return self._seq

    @seq.setter
    def seq(self, value: int):
        if not isinstance(value, int):
            raise ValueError("seq must be an integer")
        self._seq = value
        self.update_encryption_session()

    @property
    def local_seed(self) -> bytes | None:
        """Get the local seed."""
        return self._local_seed

    @local_seed.setter
    def local_seed(self, value: bytes):
        print("setting local_seed")
        if not isinstance(value, bytes):
            raise ValueError("local_seed must be bytes")
        elif len(value) != 16:
            raise ValueError("local_seed must be 16 bytes")
        else:
            self._local_seed = value
            self._remote_seed_auth_hash = None
            self._remote_seed = None
            self.update_encryption_session()

    @property
    def remote_auth_hash(self) -> bytes | None:
        """Get the remote auth hash."""
        return self._remote_seed_auth_hash

    @remote_auth_hash.setter
    def remote_auth_hash(self, value: bytes):
        print("setting remote_auth_hash")
        if not isinstance(value, bytes):
            raise ValueError("remote_auth_hash must be bytes")
        elif len(value) != 32:
            raise ValueError("remote_auth_hash must be 32 bytes")
        else:
            self._remote_seed_auth_hash = value
            self.update_encryption_session()

    @property
    def remote_seed(self) -> bytes | None:
        """Get the remote seed."""
        return self._remote_seed

    @remote_seed.setter
    def remote_seed(self, value: bytes):
        print("setting remote_seed")
        if not isinstance(value, bytes):
            raise ValueError("remote_seed must be bytes")
        elif len(value) != 16:
            raise ValueError("remote_seed must be 16 bytes")
        else:
            self._remote_seed = value
            self.update_encryption_session()

    # This function decrypts the data using the encryption session.
    def decrypt(self, *args, **kwargs):
        """Decrypt the data using the encryption session."""
        if self._session is None:
            raise ValueError("No session available")
        return self._session.decrypt(*args, **kwargs)


# This is a custom error handler that replaces bad characters with '*',
# in case something goes wrong in decryption.
# Without this, the decryption could yield an error.
def bad_chars_replacement(exception):
    """Replace bad characters with '*'."""
    return ("*", exception.start + 1)


codecs.register_error("bad_chars_replacement", bad_chars_replacement)


def main(
    loop: asyncio.AbstractEventLoop,
    username,
    password,
    device_ip,
    source_host,
    pcap_file_path,
    output_json_name=None,
):
    """Run the main function."""
    asyncio.set_event_loop(loop)
    capture = pyshark.FileCapture(pcap_file_path, display_filter="http", eventloop=loop)

    # In an effort to keep this code tied into the original code
    # (so that this can hopefully leverage any future codebase updates inheriently),
    # some weird initialization is done here
    creds = Credentials(username, password)

    fake_connection = DeviceConnectionParameters(
        DeviceFamily.SmartTapoBulb, DeviceEncryptionType.Klap
    )
    fake_device = DeviceConfig(
        device_ip, connection_type=fake_connection, credentials=creds
    )

    operator = Operator(KlapTransportV2(config=fake_device), creds)
    packets = []

    # pyshark is a little weird in how it handles iteration,
    # so this is a workaround to allow for (advanced) iteration over the packets.
    while True:
        try:
            packet = capture.next()
            packet_number = capture._current_packet
            if packet.ip.src != source_host:
                continue
            # we only care about http packets
            # this is redundant, as pyshark is set to only load http packets
            if not hasattr(packet, "http"):
                continue

            uri = packet.http.get("request_uri_path", packet.http.get("request_uri"))
            if uri is None:
                continue

            operator.seq = _get_seq_from_query(packet)

            # Windows and linux file_data attribute returns different
            # pretty format so get the raw field value.
            data = packet.http.get_field_value("file_data", raw=True)

            match uri:
                case "/app/request":
                    if packet.ip.dst != device_ip:
                        continue
                    message = bytes.fromhex(data)
                    try:
                        plaintext = operator.decrypt(message)
                        payload = json.loads(plaintext)
                        print(json.dumps(payload, indent=2))
                        packets.append(payload)
                    except ValueError:
                        print("Insufficient data to decrypt thus far")

                case "/app/handshake1":
                    if packet.ip.dst != device_ip:
                        continue
                    message = bytes.fromhex(data)
                    operator.local_seed = message
                    response = None
                    print(
                        f"got handshake1 in {packet_number}, looking for the response"
                    )
                    while (
                        True
                    ):  # we are going to now look for the response to this request
                        response = capture.next()
                        if _is_http_response_for_packet(response, packet):
                            print(f"found response in {packet_number}")
                            break
                    data = response.http.get_field_value("file_data", raw=True)
                    message = bytes.fromhex(data)
                    operator.remote_seed = message[0:16]
                    operator.remote_auth_hash = message[16:]

                case "/app/handshake2":
                    continue  # we don't care about this
                case _:
                    continue
        except StopIteration:
            break

    # save the final array to a file
    if output_json_name is not None:
        with open(output_json_name, "w") as f:
            f.write(json.dumps(packets, indent=2))
            f.write("\n" * 1)
            f.close()

    # Call close method which cleans up event loop
    capture.close()


@click.command()
@click.option(
    "--host",
    required=True,
    help="the IP of the smart device as it appears in the pcap file.",
)
@click.option(
    "--source-host",
    required=True,
    help="the IP of the device communicating with the smart device.",
)
@click.option(
    "--username",
    required=True,
    envvar="KASA_USERNAME",
    help="Username/email address to authenticate to device.",
)
@click.option(
    "--password",
    required=True,
    envvar="KASA_PASSWORD",
    help="Password to use to authenticate to device.",
)
@click.option(
    "--pcap-file-path",
    required=True,
    help="The path to the pcap file to parse.",
)
@click.option(
    "-o",
    "--output",
    required=False,
    help="The name of the output file, relative to the current directory.",
)
async def cli(username, password, host, source_host, pcap_file_path, output):
    """Export KLAP data in JSON format from a PCAP file."""
    # pyshark does not work within a running event loop and we don't want to
    # install click as well as asyncclick so run in a new thread.
    loop = asyncio.new_event_loop()
    thread = Thread(
        target=main,
        args=[loop, username, password, host, source_host, pcap_file_path, output],
        daemon=True,
    )
    thread.start()
    thread.join()


if __name__ == "__main__":
    cli()



"""Implementation of lens mask privacy module."""

from __future__ import annotations

import logging

from ...smart.smartmodule import allow_update_after
from ..smartcammodule import SmartCamModule

_LOGGER = logging.getLogger(__name__)


class LensMask(SmartCamModule):
    """Implementation of lens mask module."""

    REQUIRED_COMPONENT = "lensMask"

    QUERY_GETTER_NAME = "getLensMaskConfig"
    QUERY_MODULE_NAME = "lens_mask"
    QUERY_SECTION_NAMES = "lens_mask_info"

    @property
    def enabled(self) -> bool:
        """Return the lens mask state."""
        return self.data["lens_mask_info"]["enabled"] == "on"

    @allow_update_after
    async def set_enabled(self, enable: bool) -> dict:
        """Set the lens mask state."""
        params = {"enabled": "on" if enable else "off"}
        return await self._device._query_setter_helper(
            "setLensMaskConfig", self.QUERY_MODULE_NAME, "lens_mask_info", params
        )



"""Tests for smartcam baby cry detection module."""

from __future__ import annotations

from kasa import Device
from kasa.smartcam.smartcammodule import SmartCamModule

from ...device_fixtures import parametrize

babycrydetection = parametrize(
    "has babycry detection",
    component_filter="babyCryDetection",
    protocol_filter={"SMARTCAM"},
)


@babycrydetection
async def test_babycrydetection(dev: Device):
    """Test device babycry detection."""
    babycry = dev.modules.get(SmartCamModule.SmartCamBabyCryDetection)
    assert babycry

    bcde_feat = dev.features.get("baby_cry_detection")
    assert bcde_feat

    original_enabled = babycry.enabled

    try:
        await babycry.set_enabled(not original_enabled)
        await dev.update()
        assert babycry.enabled is not original_enabled
        assert bcde_feat.value is not original_enabled

        await babycry.set_enabled(original_enabled)
        await dev.update()
        assert babycry.enabled is original_enabled
        assert bcde_feat.value is original_enabled

        await bcde_feat.set_value(not original_enabled)
        await dev.update()
        assert babycry.enabled is not original_enabled
        assert bcde_feat.value is not original_enabled

    finally:
        await babycry.set_enabled(original_enabled)



"""Implementation of the motion detection (PIR) module found in some dimmers."""

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from enum import Enum

from ...exceptions import KasaException
from ...feature import Feature
from ..iotmodule import IotModule, merge

_LOGGER = logging.getLogger(__name__)


class Range(Enum):
    """Range for motion detection."""

    Far = 0
    Mid = 1
    Near = 2
    Custom = 3

    def __str__(self) -> str:
        return self.name


@dataclass
class PIRConfig:
    """Dataclass representing a PIR sensor configuration."""

    enabled: bool
    adc_min: int
    adc_max: int
    range: Range
    threshold: int

    @property
    def adc_mid(self) -> int:
        """Compute the ADC midpoint from the configured ADC Max and Min values."""
        return math.floor(abs(self.adc_max - self.adc_min) / 2)


@dataclass
class PIRStatus:
    """Dataclass representing the current trigger state of an ADC PIR sensor."""

    pir_config: PIRConfig
    adc_value: int

    @property
    def pir_value(self) -> int:
        """
        Get the PIR status value in integer form.

        Computes the PIR status value that this object represents,
        using the given PIR configuration.
        """
        return self.pir_config.adc_mid - self.adc_value

    @property
    def pir_percent(self) -> float:
        """
        Get the PIR status value in percentile form.

        Computes the PIR status percentage that this object represents,
        using the given PIR configuration.
        """
        value = self.pir_value
        divisor = (
            (self.pir_config.adc_mid - self.pir_config.adc_min)
            if (value < 0)
            else (self.pir_config.adc_max - self.pir_config.adc_mid)
        )
        return (float(value) / divisor) * 100

    @property
    def pir_triggered(self) -> bool:
        """
        Get the PIR status trigger state.

        Compute the PIR trigger state this object represents,
        using the given PIR configuration.
        """
        return (self.pir_config.enabled) and (
            abs(self.pir_percent) > (100 - self.pir_config.threshold)
        )


class Motion(IotModule):
    """Implements the motion detection (PIR) module."""

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        # Only add features if the device supports the module
        if "get_config" not in self.data:
            return

        # Require that ADC value is also present.
        if "get_adc_value" not in self.data:
            _LOGGER.warning("%r initialized, but no get_adc_value in response")
            return

        if "enable" not in self.config:
            _LOGGER.warning("%r initialized, but no enable in response")
            return

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_enabled",
                name="PIR enabled",
                icon="mdi:motion-sensor",
                attribute_getter="enabled",
                attribute_setter="set_enabled",
                type=Feature.Type.Switch,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_range",
                name="Motion Sensor Range",
                icon="mdi:motion-sensor",
                attribute_getter="range",
                attribute_setter="_set_range_from_str",
                type=Feature.Type.Choice,
                choices_getter="ranges",
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_threshold",
                name="Motion Sensor Threshold",
                icon="mdi:motion-sensor",
                attribute_getter="threshold",
                attribute_setter="set_threshold",
                type=Feature.Type.Number,
                category=Feature.Category.Config,
                range_getter=lambda: (0, 100),
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_triggered",
                name="PIR Triggered",
                icon="mdi:motion-sensor",
                attribute_getter="pir_triggered",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Primary,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_value",
                name="PIR Value",
                icon="mdi:motion-sensor",
                attribute_getter="pir_value",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Info,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_adc_value",
                name="PIR ADC Value",
                icon="mdi:motion-sensor",
                attribute_getter="adc_value",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_adc_min",
                name="PIR ADC Min",
                icon="mdi:motion-sensor",
                attribute_getter="adc_min",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_adc_mid",
                name="PIR ADC Mid",
                icon="mdi:motion-sensor",
                attribute_getter="adc_mid",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_adc_max",
                name="PIR ADC Max",
                icon="mdi:motion-sensor",
                attribute_getter="adc_max",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="pir_percent",
                name="PIR Percentile",
                icon="mdi:motion-sensor",
                attribute_getter="pir_percent",
                attribute_setter=None,
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
                unit_getter=lambda: "%",
            )
        )

    def query(self) -> dict:
        """Request PIR configuration."""
        req = merge(
            self.query_for_command("get_config"),
            self.query_for_command("get_adc_value"),
        )

        return req

    @property
    def config(self) -> dict:
        """Return current configuration."""
        return self.data["get_config"]

    @property
    def pir_config(self) -> PIRConfig:
        """Return PIR sensor configuration."""
        pir_range = Range(self.config["trigger_index"])
        return PIRConfig(
            enabled=bool(self.config["enable"]),
            adc_min=int(self.config["min_adc"]),
            adc_max=int(self.config["max_adc"]),
            range=pir_range,
            threshold=self.get_range_threshold(pir_range),
        )

    @property
    def enabled(self) -> bool:
        """Return True if module is enabled."""
        return self.pir_config.enabled

    @property
    def adc_min(self) -> int:
        """Return minimum ADC sensor value."""
        return self.pir_config.adc_min

    @property
    def adc_max(self) -> int:
        """Return maximum ADC sensor value."""
        return self.pir_config.adc_max

    @property
    def adc_mid(self) -> int:
        """
        Return the midpoint for the ADC.

        The midpoint represents the zero point for the PIR sensor waveform.

        Currently this is estimated by:
            math.floor(abs(adc_max - adc_min) / 2)
        """
        return self.pir_config.adc_mid

    async def set_enabled(self, state: bool) -> dict:
        """Enable/disable PIR."""
        return await self.call("set_enable", {"enable": int(state)})

    @property
    def ranges(self) -> list[str]:
        """Return set of supported range classes."""
        range_min = 0
        range_max = len(self.config["array"])
        valid_ranges = list()
        for r in Range:
            if (r.value >= range_min) and (r.value < range_max):
                valid_ranges.append(r.name)
        return valid_ranges

    @property
    def range(self) -> Range:
        """Return motion detection Range."""
        return self.pir_config.range

    async def set_range(self, range: Range) -> dict:
        """Set the Range for the sensor.

        :param Range: the range class to use.
        """
        payload = {"index": range.value}
        return await self.call("set_trigger_sens", payload)

    def _parse_range_value(self, value: str) -> Range:
        """Attempt to parse a range value from the given string."""
        value = value.strip().capitalize()
        try:
            return Range[value]
        except KeyError:
            raise KasaException(
                f"Invalid range value: '{value}'."
                f" Valid options are: {Range._member_names_}"
            ) from KeyError

    async def _set_range_from_str(self, input: str) -> dict:
        value = self._parse_range_value(input)
        return await self.set_range(range=value)

    def get_range_threshold(self, range_type: Range) -> int:
        """Get the distance threshold at which the PIR sensor is will trigger."""
        if range_type.value < 0 or range_type.value >= len(self.config["array"]):
            raise KasaException(
                "Range type is outside the bounds of the configured device ranges."
            )
        return int(self.config["array"][range_type.value])

    @property
    def threshold(self) -> int:
        """Return motion detection Range."""
        return self.pir_config.threshold

    async def set_threshold(self, value: int) -> dict:
        """Set the distance threshold at which the PIR sensor is will trigger."""
        payload = {"index": Range.Custom.value, "value": value}
        return await self.call("set_trigger_sens", payload)

    @property
    def inactivity_timeout(self) -> int:
        """Return inactivity timeout in milliseconds."""
        return self.config["cold_time"]

    async def set_inactivity_timeout(self, timeout: int) -> dict:
        """Set inactivity timeout in milliseconds.

        Note, that you need to delete the default "Smart Control" rule in the app
        to avoid reverting this back to 60 seconds after a period of time.
        """
        return await self.call("set_cold_time", {"cold_time": timeout})

    @property
    def pir_state(self) -> PIRStatus:
        """Return cached PIR status."""
        return PIRStatus(self.pir_config, self.data["get_adc_value"]["value"])

    async def get_pir_state(self) -> PIRStatus:
        """Return real-time PIR status."""
        latest = await self.call("get_adc_value")
        self.data["get_adc_value"] = latest
        return PIRStatus(self.pir_config, latest["value"])

    @property
    def adc_value(self) -> int:
        """Return motion adc value."""
        return self.pir_state.adc_value

    @property
    def pir_value(self) -> int:
        """Return the computed PIR sensor value."""
        return self.pir_state.pir_value

    @property
    def pir_percent(self) -> float:
        """Return the computed PIR sensor value, in percentile form."""
        return self.pir_state.pir_percent

    @property
    def pir_triggered(self) -> bool:
        """Return if the motion sensor has been triggered."""
        return self.pir_state.pir_triggered



"""Implementation of temperature module."""

from __future__ import annotations

from typing import Literal

from ...feature import Feature
from ..smartmodule import SmartModule


class TemperatureSensor(SmartModule):
    """Implementation of temperature module."""

    REQUIRED_COMPONENT = "temperature"
    QUERY_GETTER_NAME = "get_comfort_temp_config"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="temperature",
                name="Temperature",
                container=self,
                attribute_getter="temperature",
                icon="mdi:thermometer",
                category=Feature.Category.Primary,
                unit_getter="temperature_unit",
                type=Feature.Type.Sensor,
            )
        )
        if "current_temp_exception" in self._device.sys_info:
            self._add_feature(
                Feature(
                    self._device,
                    id="temperature_warning",
                    name="Temperature warning",
                    container=self,
                    attribute_getter="temperature_warning",
                    type=Feature.Type.BinarySensor,
                    icon="mdi:alert",
                    category=Feature.Category.Debug,
                )
            )
        self._add_feature(
            Feature(
                self._device,
                id="temperature_unit",
                name="Temperature unit",
                container=self,
                attribute_getter="temperature_unit",
                attribute_setter="set_temperature_unit",
                type=Feature.Type.Choice,
                choices_getter=lambda: ["celsius", "fahrenheit"],
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    @property
    def temperature(self) -> float:
        """Return current humidity in percentage."""
        return self._device.sys_info["current_temp"]

    @property
    def temperature_warning(self) -> bool:
        """Return True if temperature is outside of the wanted range."""
        return self._device.sys_info.get("current_temp_exception", 0) != 0

    @property
    def temperature_unit(self) -> Literal["celsius", "fahrenheit"]:
        """Return current temperature unit."""
        return self._device.sys_info["temp_unit"]

    async def set_temperature_unit(
        self, unit: Literal["celsius", "fahrenheit"]
    ) -> dict:
        """Set the device temperature unit."""
        return await self.call("set_temperature_unit", {"temp_unit": unit})



from pytest_mock import MockerFixture

from kasa import Module
from kasa.iot import IotDimmer
from kasa.iot.modules.ambientlight import AmbientLight

from ...device_fixtures import dimmer_iot


@dimmer_iot
def test_ambientlight_getters(dev: IotDimmer):
    assert Module.IotAmbientLight in dev.modules
    ambientlight: AmbientLight = dev.modules[Module.IotAmbientLight]

    assert ambientlight.enabled == ambientlight.config["enable"]
    assert ambientlight.presets == ambientlight.config["level_array"]

    assert (
        ambientlight.ambientlight_brightness
        == ambientlight.data["get_current_brt"]["value"]
    )


@dimmer_iot
async def test_ambientlight_setters(dev: IotDimmer, mocker: MockerFixture):
    ambientlight: AmbientLight = dev.modules[Module.IotAmbientLight]
    query_helper = mocker.patch("kasa.iot.IotDimmer._query_helper")

    await ambientlight.set_enabled(True)
    query_helper.assert_called_with("smartlife.iot.LAS", "set_enable", {"enable": True})

    await ambientlight.set_brightness_limit(10)
    query_helper.assert_called_with(
        "smartlife.iot.LAS", "set_brt_level", {"index": 0, "value": 10}
    )


@dimmer_iot
def test_ambientlight_feature(dev: IotDimmer):
    assert Module.IotAmbientLight in dev.modules
    ambientlight: AmbientLight = dev.modules[Module.IotAmbientLight]

    enabled = dev.features["ambient_light_enabled"]
    assert ambientlight.enabled == enabled.value

    brightness = dev.features["ambient_light"]
    assert ambientlight.ambientlight_brightness == brightness.value



"""Implementation of alarm module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, TypeAlias

from ...feature import Feature
from ...interfaces import Alarm as AlarmInterface
from ...module import FeatureAttribute
from ..smartmodule import SmartModule

DURATION_MAX = 10 * 60

VOLUME_INT_TO_STR = {
    0: "mute",
    1: "low",
    2: "normal",
    3: "high",
}

VOLUME_STR_LIST = [v for v in VOLUME_INT_TO_STR.values()]
VOLUME_INT_RANGE = (min(VOLUME_INT_TO_STR.keys()), max(VOLUME_INT_TO_STR.keys()))
VOLUME_STR_TO_INT = {v: k for k, v in VOLUME_INT_TO_STR.items()}

AlarmVolume: TypeAlias = Literal["mute", "low", "normal", "high"]


class Alarm(SmartModule, AlarmInterface):
    """Implementation of alarm module."""

    REQUIRED_COMPONENT = "alarm"

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {
            "get_alarm_configure": None,
            "get_support_alarm_type_list": None,  # This should be needed only once
        }

    def _initialize_features(self) -> None:
        """Initialize features."""
        device = self._device
        self._add_feature(
            Feature(
                device,
                id="alarm",
                name="Alarm",
                container=self,
                attribute_getter="active",
                icon="mdi:bell",
                type=Feature.Type.BinarySensor,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="alarm_source",
                name="Alarm source",
                container=self,
                attribute_getter="source",
                icon="mdi:bell",
                type=Feature.Type.Sensor,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="alarm_sound",
                name="Alarm sound",
                container=self,
                attribute_getter="alarm_sound",
                attribute_setter="set_alarm_sound",
                category=Feature.Category.Config,
                type=Feature.Type.Choice,
                choices_getter="alarm_sounds",
            )
        )
        self._add_feature(
            Feature(
                device,
                id="alarm_volume",
                name="Alarm volume",
                container=self,
                attribute_getter="_alarm_volume_str",
                attribute_setter="set_alarm_volume",
                category=Feature.Category.Config,
                type=Feature.Type.Choice,
                choices_getter=lambda: VOLUME_STR_LIST,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="alarm_volume_level",
                name="Alarm volume",
                container=self,
                attribute_getter="alarm_volume",
                attribute_setter="set_alarm_volume",
                category=Feature.Category.Config,
                type=Feature.Type.Number,
                range_getter=lambda: VOLUME_INT_RANGE,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="alarm_duration",
                name="Alarm duration",
                container=self,
                attribute_getter="alarm_duration",
                attribute_setter="set_alarm_duration",
                category=Feature.Category.Config,
                type=Feature.Type.Number,
                range_getter=lambda: (1, DURATION_MAX),
            )
        )
        self._add_feature(
            Feature(
                device,
                id="test_alarm",
                name="Test alarm",
                container=self,
                attribute_setter="play",
                type=Feature.Type.Action,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="stop_alarm",
                name="Stop alarm",
                container=self,
                attribute_setter="stop",
                type=Feature.Type.Action,
            )
        )

    @property
    def alarm_sound(self) -> Annotated[str, FeatureAttribute()]:
        """Return current alarm sound."""
        return self.data["get_alarm_configure"]["type"]

    async def set_alarm_sound(self, sound: str) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm sound.

        See *alarm_sounds* for list of available sounds.
        """
        self._check_sound(sound)
        payload = self.data["get_alarm_configure"].copy()
        payload["type"] = sound
        return await self.call("set_alarm_configure", payload)

    @property
    def alarm_sounds(self) -> list[str]:
        """Return list of available alarm sounds."""
        return self.data["get_support_alarm_type_list"]["alarm_type_list"]

    @property
    def alarm_volume(self) -> Annotated[int, FeatureAttribute("alarm_volume_level")]:
        """Return alarm volume."""
        return VOLUME_STR_TO_INT[self._alarm_volume_str]

    @property
    def _alarm_volume_str(
        self,
    ) -> Annotated[AlarmVolume, FeatureAttribute("alarm_volume")]:
        """Return alarm volume."""
        return self.data["get_alarm_configure"]["volume"]

    async def set_alarm_volume(
        self, volume: AlarmVolume | int
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm volume."""
        self._check_and_convert_volume(volume)
        payload = self.data["get_alarm_configure"].copy()
        payload["volume"] = volume
        return await self.call("set_alarm_configure", payload)

    @property
    def alarm_duration(self) -> Annotated[int, FeatureAttribute()]:
        """Return alarm duration."""
        return self.data["get_alarm_configure"]["duration"]

    async def set_alarm_duration(
        self, duration: int
    ) -> Annotated[dict, FeatureAttribute()]:
        """Set alarm duration."""
        self._check_duration(duration)
        payload = self.data["get_alarm_configure"].copy()
        payload["duration"] = duration
        return await self.call("set_alarm_configure", payload)

    @property
    def active(self) -> bool:
        """Return true if alarm is active."""
        return self._device.sys_info["in_alarm"]

    @property
    def source(self) -> str | None:
        """Return the alarm cause."""
        src = self._device.sys_info["in_alarm_source"]
        return src if src else None

    async def play(
        self,
        *,
        duration: int | None = None,
        volume: int | AlarmVolume | None = None,
        sound: str | None = None,
    ) -> dict:
        """Play alarm.

        The optional *duration*, *volume*, and *sound* to override the device settings.
        *volume* can be set to 'mute', 'low', 'normal', or 'high'.
        *duration* is in seconds.
        See *alarm_sounds* for the list of sounds available for the device.
        """
        params: dict[str, str | int] = {}

        if duration is not None:
            self._check_duration(duration)
            params["alarm_duration"] = duration

        if volume is not None:
            target_volume = self._check_and_convert_volume(volume)
            params["alarm_volume"] = target_volume

        if sound is not None:
            self._check_sound(sound)
            params["alarm_type"] = sound

        return await self.call("play_alarm", params)

    async def stop(self) -> dict:
        """Stop alarm."""
        return await self.call("stop_alarm")

    def _check_and_convert_volume(self, volume: str | int) -> str:
        """Raise an exception on invalid volume."""
        if isinstance(volume, int):
            volume = VOLUME_INT_TO_STR.get(volume, "invalid")

        if TYPE_CHECKING:
            assert isinstance(volume, str)

        if volume not in VOLUME_INT_TO_STR.values():
            raise ValueError(
                f"Invalid volume {volume} "
                f"available: {VOLUME_INT_TO_STR.keys()}, {VOLUME_INT_TO_STR.values()}"
            )

        return volume

    def _check_duration(self, duration: int) -> None:
        """Raise an exception on invalid duration."""
        if duration < 1 or duration > DURATION_MAX:
            raise ValueError(f"Invalid duration {duration} available: 1-600")

    def _check_sound(self, sound: str) -> None:
        """Raise an exception on invalid sound."""
        if sound not in self.alarm_sounds:
            raise ValueError(f"Invalid sound {sound} available: {self.alarm_sounds}")



"""Tests for SMART devices."""

from __future__ import annotations

import copy
import logging
import time
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, cast
from unittest.mock import patch

import pytest
from freezegun.api import FrozenDateTimeFactory
from pytest_mock import MockerFixture

from kasa import Device, DeviceType, KasaException, Module
from kasa.exceptions import DeviceError, SmartErrorCode
from kasa.smart import SmartDevice
from kasa.smart.modules.energy import Energy
from kasa.smart.smartmodule import SmartModule
from kasa.smartcam import SmartCamDevice
from tests.conftest import (
    DISCOVERY_MOCK_IP,
    device_smart,
    get_device_for_fixture_protocol,
    get_parent_and_child_modules,
    smart_discovery,
)
from tests.device_fixtures import (
    hub_smartcam,
    hubs_smart,
    parametrize_combine,
    variable_temp_smart,
)

from ..fakeprotocol_smart import FakeSmartTransport
from ..fakeprotocol_smartcam import FakeSmartCamTransport

DUMMY_CHILD_REQUEST_PREFIX = "get_dummy_"

hub_all = parametrize_combine([hubs_smart, hub_smartcam])


@device_smart
@pytest.mark.requires_dummy
@pytest.mark.xdist_group(name="caplog")
async def test_try_get_response(dev: SmartDevice, caplog):
    mock_response: dict = {
        "get_device_info": SmartErrorCode.PARAMS_ERROR,
    }
    caplog.set_level(logging.DEBUG)
    dev._try_get_response(mock_response, "get_device_info", {})
    msg = "Error PARAMS_ERROR(-1008) getting request get_device_info for device 127.0.0.123"
    assert msg in caplog.text


@device_smart
@pytest.mark.requires_dummy
async def test_update_no_device_info(dev: SmartDevice, mocker: MockerFixture):
    mock_response: dict = {
        "get_device_usage": {},
        "get_device_time": {},
    }
    msg = f"get_device_info not found in {mock_response} for device 127.0.0.123"
    mocker.patch.object(dev.protocol, "query", return_value=mock_response)
    with pytest.raises(KasaException, match=msg):
        await dev.update()


@smart_discovery
async def test_device_type_no_update(discovery_mock, caplog: pytest.LogCaptureFixture):
    """Test device type and repr when device not updated."""
    dev = SmartDevice(DISCOVERY_MOCK_IP)
    assert dev.device_type is DeviceType.Unknown
    assert repr(dev) == f"<DeviceType.Unknown at {DISCOVERY_MOCK_IP} - update() needed>"

    discovery_result = copy.deepcopy(discovery_mock.discovery_data["result"])

    disco_model = discovery_result["device_model"]
    short_model, _, _ = disco_model.partition("(")
    dev.update_from_discover_info(discovery_result)
    assert dev.device_type is DeviceType.Unknown
    assert (
        repr(dev)
        == f"<DeviceType.Unknown at {DISCOVERY_MOCK_IP} - None ({short_model}) - update() needed>"
    )
    discovery_result["device_type"] = "SMART.FOOBAR"
    dev.update_from_discover_info(discovery_result)
    dev._components = {"dummy": 1}
    assert dev.device_type is DeviceType.Plug
    assert (
        repr(dev)
        == f"<DeviceType.Plug at {DISCOVERY_MOCK_IP} - None ({short_model}) - update() needed>"
    )
    assert "Unknown device type, falling back to plug" in caplog.text


@device_smart
async def test_initial_update(dev: SmartDevice, mocker: MockerFixture):
    """Test the initial update cycle."""
    # As the fixture data is already initialized, we reset the state for testing
    dev._components_raw = None
    dev._components = {}
    dev._modules = OrderedDict()
    dev._features = {}
    dev._children = {}
    dev._last_update = {}
    dev._last_update_time = None

    negotiate = mocker.spy(dev, "_negotiate")
    initialize_modules = mocker.spy(dev, "_initialize_modules")
    initialize_features = mocker.spy(dev, "_initialize_features")

    # Perform two updates and verify that initialization is only done once
    await dev.update()
    await dev.update()

    negotiate.assert_called_once()
    assert dev._components_raw is not None
    initialize_modules.assert_called_once()
    assert dev.modules
    initialize_features.assert_called_once()
    assert dev.features


@device_smart
async def test_negotiate(dev: SmartDevice, mocker: MockerFixture):
    """Test that the initial negotiation performs expected steps."""
    # As the fixture data is already initialized, we reset the state for testing
    dev._components_raw = None
    dev._children = {}

    query = mocker.spy(dev.protocol, "query")
    initialize_children = mocker.spy(dev, "_initialize_children")
    await dev._negotiate()

    # Check that we got the initial negotiation call
    query.assert_any_call(
        {
            "component_nego": None,
            "get_device_info": None,
            "get_connect_cloud_state": None,
        }
    )
    assert dev._components_raw

    # Check the children are created, if device supports them
    if "child_device" in dev._components:
        initialize_children.assert_called_once()
        query.assert_any_call(
            {
                "get_child_device_component_list": None,
                "get_child_device_list": None,
            }
        )
        await dev.update()
        assert len(dev._children) == dev.internal_state["get_child_device_list"]["sum"]


@device_smart
async def test_update_module_queries(dev: SmartDevice, mocker: MockerFixture):
    """Test that the regular update uses queries from all supported modules."""
    # We need to have some modules initialized by now
    assert dev._modules
    # Reset last update so all modules will query
    for mod in dev._modules.values():
        mod._last_update_time = None

    device_queries: dict[SmartDevice, dict[str, Any]] = {}
    for mod in dev._modules.values():
        device_queries.setdefault(mod._device, {}).update(mod.query())
    # Hubs do not query child modules by default.
    if dev.device_type != Device.Type.Hub:
        for child in dev.children:
            for mod in child.modules.values():
                device_queries.setdefault(mod._device, {}).update(mod.query())

    spies = {}
    for device in device_queries:
        spies[device] = mocker.spy(device.protocol, "query")

    await dev.update()
    for device in device_queries:
        if device_queries[device]:
            # Need assert any here because the child device updates use the parent's protocol
            spies[device].assert_any_call(device_queries[device])
        else:
            spies[device].assert_not_called()


@device_smart
@pytest.mark.xdist_group(name="caplog")
async def test_update_module_update_delays(
    dev: SmartDevice,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    freezer: FrozenDateTimeFactory,
):
    """Test that modules with minimum delays delay."""
    # We need to have some modules initialized by now
    assert dev._modules

    new_dev = SmartDevice("127.0.0.1", protocol=dev.protocol)
    await new_dev.update()
    first_update_time = time.monotonic()
    assert new_dev._last_update_time == first_update_time
    for module in new_dev.modules.values():
        if module.query():
            assert module._last_update_time == first_update_time

    seconds = 0
    tick = 30
    while seconds <= 180:
        seconds += tick
        freezer.tick(tick)

        now = time.monotonic()
        await new_dev.update()
        for module in new_dev.modules.values():
            mod_delay = module.MINIMUM_UPDATE_INTERVAL_SECS
            if module.query():
                expected_update_time = (
                    now if mod_delay == 0 else now - (seconds % mod_delay)
                )

                assert module._last_update_time == expected_update_time, (
                    f"Expected update time {expected_update_time} after {seconds} seconds for {module.name} with delay {mod_delay} got {module._last_update_time}"
                )


async def _get_child_responses(child_requests: list[dict[str, Any]], child_protocol):
    """Get dummy responses for testing all child modules.

    Even if they don't return really return query.
    """
    child_req = {item["method"]: item.get("params") for item in child_requests}
    child_resp = {k: v for k, v in child_req.items() if k.startswith("get_dummy")}
    child_req = {
        k: v for k, v in child_req.items() if k.startswith("get_dummy") is False
    }
    resp = await child_protocol._query(child_req)
    resp = {**child_resp, **resp}
    return [
        {"method": k, "error_code": 0, "result": v or {"dummy": "dummy"}}
        for k, v in resp.items()
    ]


@hub_all
@pytest.mark.xdist_group(name="caplog")
async def test_hub_children_update_delays(
    dev: SmartDevice,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    freezer: FrozenDateTimeFactory,
):
    """Test that hub children use the correct delay."""
    if not dev.children:
        pytest.skip(f"Device {dev.model} does not have children.")
    # We need to have some modules initialized by now
    assert dev._modules

    new_dev = type(dev)("127.0.0.1", protocol=dev.protocol)
    module_queries: dict[str, dict[str, dict]] = {}

    # children should always update on first update
    await new_dev.update(update_children=False)

    if TYPE_CHECKING:
        from ..fakeprotocol_smart import FakeSmartTransport

        assert isinstance(dev.protocol._transport, FakeSmartTransport)
    if dev.protocol._transport.child_protocols:
        for child in new_dev.children:
            for modname, module in child._modules.items():
                if (
                    not (q := module.query())
                    and modname not in {"DeviceModule", "Light", "Battery", "Camera"}
                    and not module.SYSINFO_LOOKUP_KEYS
                ):
                    q = {f"get_dummy_{modname}": {}}
                    mocker.patch.object(module, "query", return_value=q)
                if q:
                    queries = module_queries.setdefault(child.device_id, {})
                    queries[cast(str, modname)] = q
                module._last_update_time = None

    module_queries[""] = {
        cast(str, modname): q
        for modname, module in dev._modules.items()
        if (q := module.query())
    }

    async def _query(request, *args, **kwargs):
        # If this is a child multipleRequest query return the error wrapped
        child_id = None
        # smart hub
        if (
            (cc := request.get("control_child"))
            and (child_id := cc.get("device_id"))
            and (requestData := cc["requestData"])
            and requestData["method"] == "multipleRequest"
            and (child_requests := requestData["params"]["requests"])
        ):
            child_protocol = dev.protocol._transport.child_protocols[child_id]
            resp = await _get_child_responses(child_requests, child_protocol)
            return {"control_child": {"responseData": {"result": {"responses": resp}}}}
        # smartcam hub
        if (
            (mr := request.get("multipleRequest"))
            and (requests := mr.get("requests"))
            # assumes all requests for the same child
            and (
                child_id := next(iter(requests))
                .get("params", {})
                .get("childControl", {})
                .get("device_id")
            )
            and (
                child_requests := [
                    cc["request_data"]
                    for req in requests
                    if (cc := req["params"].get("childControl"))
                ]
            )
        ):
            child_protocol = dev.protocol._transport.child_protocols[child_id]
            resp = await _get_child_responses(child_requests, child_protocol)
            resp = [{"result": {"response_data": resp}} for resp in resp]
            return {"multipleRequest": {"responses": resp}}

        if child_id:  # child single query
            child_protocol = dev.protocol._transport.child_protocols[child_id]
            resp_list = await _get_child_responses([requestData], child_protocol)
            resp = {"control_child": {"responseData": resp_list[0]}}
        else:
            resp = await dev.protocol._query(request, *args, **kwargs)

        return resp

    mocker.patch.object(new_dev.protocol, "query", side_effect=_query)

    first_update_time = time.monotonic()
    assert new_dev._last_update_time == first_update_time

    await new_dev.update()

    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            mod = cast(SmartModule, check_dev.modules[modname])
            assert mod._last_update_time == first_update_time

    for mod in new_dev.modules.values():
        mod.MINIMUM_UPDATE_INTERVAL_SECS = 5
    freezer.tick(180)

    now = time.monotonic()
    await new_dev.update()

    child_tick = max(
        module.MINIMUM_HUB_CHILD_UPDATE_INTERVAL_SECS
        for child in new_dev.children
        for module in child.modules.values()
    )

    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            if modname in {"Firmware"}:
                continue
            mod = cast(SmartModule, check_dev.modules[modname])
            expected_update_time = first_update_time if dev_id else now
            assert mod._last_update_time == expected_update_time

    freezer.tick(child_tick)

    now = time.monotonic()
    await new_dev.update()

    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            if modname in {"Firmware"}:
                continue
            mod = cast(SmartModule, check_dev.modules[modname])

            assert mod._last_update_time == now


@pytest.mark.parametrize(
    ("first_update"),
    [
        pytest.param(True, id="First update true"),
        pytest.param(False, id="First update false"),
    ],
)
@pytest.mark.parametrize(
    ("error_type"),
    [
        pytest.param(SmartErrorCode.PARAMS_ERROR, id="Device error"),
        pytest.param(TimeoutError("Dummy timeout"), id="Query error"),
    ],
)
@pytest.mark.parametrize(
    ("recover"),
    [
        pytest.param(True, id="recover"),
        pytest.param(False, id="no recover"),
    ],
)
@device_smart
@pytest.mark.xdist_group(name="caplog")
async def test_update_module_query_errors(
    dev: SmartDevice,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    freezer: FrozenDateTimeFactory,
    first_update,
    error_type,
    recover,
):
    """Test that modules that disabled / removed on query failures.

    i.e. the whole query times out rather than device returns an error.
    """
    # We need to have some modules initialized by now
    assert dev._modules

    SmartModule.DISABLE_AFTER_ERROR_COUNT = 2
    first_update_queries = {"get_device_info", "get_connect_cloud_state"}

    critical_modules = {Module.DeviceModule, Module.ChildDevice}

    new_dev = SmartDevice("127.0.0.1", protocol=dev.protocol)
    if not first_update:
        await new_dev.update()
        freezer.tick(max(module.update_interval for module in dev._modules.values()))

    module_queries: dict[str, dict[str, dict]] = {}
    if TYPE_CHECKING:
        from ..fakeprotocol_smart import FakeSmartTransport

        assert isinstance(dev.protocol._transport, FakeSmartTransport)
    if dev.protocol._transport.child_protocols:
        for child in new_dev.children:
            for modname, module in child._modules.items():
                if (
                    not (q := module.query())
                    and modname not in {"DeviceModule", "Light"}
                    and not module.SYSINFO_LOOKUP_KEYS
                ):
                    q = {f"get_dummy_{modname}": {}}
                    mocker.patch.object(module, "query", return_value=q)
                if q:
                    queries = module_queries.setdefault(child.device_id, {})
                    queries[cast(str, modname)] = q

    module_queries[""] = {
        cast(str, modname): q
        for modname, module in dev._modules.items()
        if (q := module.query()) and modname not in critical_modules
    }

    raise_error = True

    async def _query(request, *args, **kwargs):
        pass
        # If this is a childmultipleRequest query return the error wrapped
        child_id = None
        if (
            (cc := request.get("control_child"))
            and (child_id := cc.get("device_id"))
            and (requestData := cc["requestData"])
            and requestData["method"] == "multipleRequest"
            and (child_requests := requestData["params"]["requests"])
        ):
            if raise_error:
                if not isinstance(error_type, SmartErrorCode):
                    raise TimeoutError()
                if len(child_requests) > 1:
                    raise TimeoutError()

            if raise_error:
                resp = {
                    "method": child_requests[0]["method"],
                    "error_code": error_type.value,
                }
            else:
                child_protocol = dev.protocol._transport.child_protocols[child_id]
                resp = await _get_child_responses(child_requests, child_protocol)
            return {"control_child": {"responseData": {"result": {"responses": resp}}}}

        if (
            not raise_error
            or "component_nego" in request
            # allow the initial child device query
            or (
                "get_child_device_component_list" in request
                and "get_child_device_list" in request
                and len(request) == 2
            )
        ):
            if child_id:  # child single query
                child_protocol = dev.protocol._transport.child_protocols[child_id]
                resp_list = await _get_child_responses([requestData], child_protocol)
                resp = {"control_child": {"responseData": resp_list[0]}}
            else:
                resp = await dev.protocol._query(request, *args, **kwargs)
            if raise_error:
                resp["get_connect_cloud_state"] = SmartErrorCode.CLOUD_FAILED_ERROR
            return resp

        # Don't test for errors on get_device_info as that is likely terminal
        if len(request) == 1 and "get_device_info" in request:
            return await dev.protocol._query(request, *args, **kwargs)

        if isinstance(error_type, SmartErrorCode):
            if len(request) == 1:
                raise DeviceError("Dummy device error", error_code=error_type)
            raise TimeoutError("Dummy timeout")
        raise error_type

    mocker.patch.object(new_dev.protocol, "query", side_effect=_query)

    await new_dev.update()

    msg = f"Error querying {new_dev.host} for modules"
    assert msg in caplog.text
    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            mod = cast(SmartModule, check_dev.modules[modname])
            if modname in {"DeviceModule"} or (
                hasattr(mod, "_state_in_sysinfo") and mod._state_in_sysinfo is True
            ):
                continue
            assert mod.disabled is False, f"{modname} disabled"
            assert mod.update_interval == mod.UPDATE_INTERVAL_AFTER_ERROR_SECS
            for mod_query in modqueries[modname]:
                if not first_update or mod_query not in first_update_queries:
                    msg = f"Error querying {new_dev.host} individually for module query '{mod_query}"
                    assert msg in caplog.text

    # Query again should not run for the modules
    caplog.clear()
    await new_dev.update()
    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            mod = cast(SmartModule, check_dev.modules[modname])
            assert mod.disabled is False, f"{modname} disabled"

    freezer.tick(SmartModule.UPDATE_INTERVAL_AFTER_ERROR_SECS)

    caplog.clear()

    if recover:
        raise_error = False

    await new_dev.update()
    msg = f"Error querying {new_dev.host} for modules"
    if not recover:
        assert msg in caplog.text

    for dev_id, modqueries in module_queries.items():
        check_dev = new_dev._children[dev_id] if dev_id else new_dev
        for modname in modqueries:
            mod = cast(SmartModule, check_dev.modules[modname])
            if modname in {"DeviceModule"} or (
                hasattr(mod, "_state_in_sysinfo") and mod._state_in_sysinfo is True
            ):
                continue
            if not recover:
                assert mod.disabled is True, f"{modname} not disabled"
                assert mod._error_count == 2
                assert mod._last_update_error
                for mod_query in modqueries[modname]:
                    if not first_update or mod_query not in first_update_queries:
                        msg = f"Error querying {new_dev.host} individually for module query '{mod_query}"
                        assert msg in caplog.text
                # Test one of the raise_if_update_error
                if mod.name == "Energy":
                    emod = cast(Energy, mod)
                    with pytest.raises(KasaException, match="Module update error"):
                        assert emod.status is not None
            else:
                assert mod.disabled is False
                assert mod._error_count == 0
                assert mod._last_update_error is None
                # Test one of the raise_if_update_error doesn't raise
                if mod.name == "Energy":
                    emod = cast(Energy, mod)
                    assert emod.status is not None


async def test_get_modules():
    """Test getting modules for child and parent modules."""
    dummy_device = await get_device_for_fixture_protocol(
        "KS240(US)_1.0_1.0.5.json", "SMART"
    )
    from kasa.smart.modules import Cloud

    # Modules on device
    module = dummy_device.modules.get("Cloud")
    assert module
    assert module.device == dummy_device
    assert isinstance(module, Cloud)

    module = dummy_device.modules.get(Module.Cloud)
    assert module
    assert module._device == dummy_device
    assert isinstance(module, Cloud)

    # Modules on child
    module = dummy_device.modules.get("Fan")
    assert module is None
    module = next(get_parent_and_child_modules(dummy_device, "Fan"))
    assert module
    assert module.device != dummy_device
    assert module.device.parent == dummy_device

    # Invalid modules
    module = dummy_device.modules.get("DummyModule")
    assert module is None

    module = dummy_device.modules.get(Module.IotAmbientLight)
    assert module is None


@device_smart
async def test_smartdevice_cloud_connection(dev: SmartDevice, mocker: MockerFixture):
    """Test is_cloud_connected property."""
    assert isinstance(dev, SmartDevice)
    assert "cloud_connect" in dev._components

    is_connected = (
        (cc := dev._last_update.get("get_connect_cloud_state"))
        and not isinstance(cc, SmartErrorCode)
        and cc["status"] == 0
    )

    assert dev.is_cloud_connected == is_connected
    last_update = dev._last_update

    for child in dev.children:
        mocker.patch.object(child.protocol, "query", return_value=child._last_update)

    last_update["get_connect_cloud_state"] = {"status": 0}
    with patch.object(dev.protocol, "query", return_value=last_update):
        await dev.update()
        assert dev.is_cloud_connected is True

    last_update["get_connect_cloud_state"] = {"status": 1}
    with patch.object(dev.protocol, "query", return_value=last_update):
        await dev.update()
        assert dev.is_cloud_connected is False

    last_update["get_connect_cloud_state"] = SmartErrorCode.UNKNOWN_METHOD_ERROR
    with patch.object(dev.protocol, "query", return_value=last_update):
        await dev.update()
        assert dev.is_cloud_connected is False

    # Test for no cloud_connect component during device initialisation
    component_list = [
        val
        for val in dev._components_raw["component_list"]
        if val["id"] not in {"cloud_connect"}
    ]
    initial_response = {
        "component_nego": {"component_list": component_list},
        "get_connect_cloud_state": last_update["get_connect_cloud_state"],
        "get_device_info": last_update["get_device_info"],
    }

    new_dev = SmartDevice("127.0.0.1", protocol=dev.protocol)

    first_call = True

    async def side_effect_func(*args, **kwargs):
        nonlocal first_call
        resp = (
            initial_response
            if first_call
            else await new_dev.protocol._query(*args, **kwargs)
        )
        first_call = False
        return resp

    with patch.object(
        new_dev.protocol,
        "query",
        side_effect=side_effect_func,
    ):
        await new_dev.update()
        assert new_dev.is_cloud_connected is False


@variable_temp_smart
async def test_smart_temp_range(dev: Device):
    light = dev.modules.get(Module.Light)
    assert light
    color_temp_feat = light.get_feature("color_temp")
    assert color_temp_feat
    assert color_temp_feat.range


@device_smart
async def test_initialize_modules_sysinfo_lookup_keys(
    dev: SmartDevice, mocker: MockerFixture
):
    """Test that matching modules using SYSINFO_LOOKUP_KEYS are initialized correctly."""

    class AvailableKey(SmartModule):
        SYSINFO_LOOKUP_KEYS = ["device_id"]

    class NonExistingKey(SmartModule):
        SYSINFO_LOOKUP_KEYS = ["this_does_not_exist"]

    # The __init_subclass__ hook in smartmodule checks the path,
    # so we have to manually add these for testing.
    mocker.patch.dict(
        "kasa.smart.smartmodule.SmartModule.REGISTERED_MODULES",
        {
            AvailableKey._module_name(): AvailableKey,
            NonExistingKey._module_name(): NonExistingKey,
        },
    )

    # We have an already initialized device, so we try to initialize the modules again
    await dev._initialize_modules()

    assert "AvailableKey" in dev.modules
    assert "NonExistingKey" not in dev.modules


@device_smart
async def test_initialize_modules_required_component(
    dev: SmartDevice, mocker: MockerFixture
):
    """Test that matching modules using REQUIRED_COMPONENT are initialized correctly."""

    class AvailableComponent(SmartModule):
        REQUIRED_COMPONENT = "device"

    class NonExistingComponent(SmartModule):
        REQUIRED_COMPONENT = "this_does_not_exist"

    # The __init_subclass__ hook in smartmodule checks the path,
    # so we have to manually add these for testing.
    mocker.patch.dict(
        "kasa.smart.smartmodule.SmartModule.REGISTERED_MODULES",
        {
            AvailableComponent._module_name(): AvailableComponent,
            NonExistingComponent._module_name(): NonExistingComponent,
        },
    )

    # We have an already initialized device, so we try to initialize the modules again
    await dev._initialize_modules()

    assert "AvailableComponent" in dev.modules
    assert "NonExistingComponent" not in dev.modules


async def test_smartmodule_query():
    """Test that a module that doesn't set QUERY_GETTER_NAME has empty query."""

    class DummyModule(SmartModule):
        pass

    dummy_device = await get_device_for_fixture_protocol(
        "KS240(US)_1.0_1.0.5.json", "SMART"
    )
    mod = DummyModule(dummy_device, "dummy")
    assert mod.query() == {}


@hub_all
@pytest.mark.xdist_group(name="caplog")
@pytest.mark.requires_dummy
async def test_dynamic_devices(dev: Device, caplog: pytest.LogCaptureFixture):
    """Test dynamic child devices."""
    if not dev.children:
        pytest.skip(f"Device {dev.model} does not have children.")

    transport = dev.protocol._transport
    assert isinstance(transport, FakeSmartCamTransport | FakeSmartTransport)

    lu = dev._last_update
    assert lu
    child_device_info = lu.get("getChildDeviceList", lu.get("get_child_device_list"))
    assert child_device_info

    child_device_components = lu.get(
        "getChildDeviceComponentList", lu.get("get_child_device_component_list")
    )
    assert child_device_components

    mock_child_device_info = copy.deepcopy(child_device_info)
    mock_child_device_components = copy.deepcopy(child_device_components)

    first_child = child_device_info["child_device_list"][0]
    first_child_device_id = first_child["device_id"]

    first_child_components = next(
        iter(
            [
                cc
                for cc in child_device_components["child_component_list"]
                if cc["device_id"] == first_child_device_id
            ]
        )
    )

    first_child_fake_transport = transport.child_protocols[first_child_device_id]

    # Test adding devices
    start_child_count = len(dev.children)
    added_ids = []
    for i in range(1, 3):
        new_child = copy.deepcopy(first_child)
        new_child_components = copy.deepcopy(first_child_components)

        mock_device_id = f"mock_child_device_id_{i}"

        transport.child_protocols[mock_device_id] = first_child_fake_transport
        new_child["device_id"] = mock_device_id
        new_child_components["device_id"] = mock_device_id

        added_ids.append(mock_device_id)
        mock_child_device_info["child_device_list"].append(new_child)
        mock_child_device_components["child_component_list"].append(
            new_child_components
        )

    def mock_get_child_device_queries(method, params):
        if method in {"getChildDeviceList", "get_child_device_list"}:
            result = mock_child_device_info
        if method in {"getChildDeviceComponentList", "get_child_device_component_list"}:
            result = mock_child_device_components
        return {"result": result, "error_code": 0}

    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    for added_id in added_ids:
        assert added_id in dev._children
    expected_new_length = start_child_count + len(added_ids)
    assert len(dev.children) == expected_new_length

    # Test removing devices
    mock_child_device_info["child_device_list"] = [
        info
        for info in mock_child_device_info["child_device_list"]
        if info["device_id"] != first_child_device_id
    ]
    mock_child_device_components["child_component_list"] = [
        cc
        for cc in mock_child_device_components["child_component_list"]
        if cc["device_id"] != first_child_device_id
    ]

    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    expected_new_length -= 1
    assert len(dev.children) == expected_new_length

    # Test no child devices

    mock_child_device_info["child_device_list"] = []
    mock_child_device_components["child_component_list"] = []
    mock_child_device_info["sum"] = 0
    mock_child_device_components["sum"] = 0

    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert len(dev.children) == 0

    # Logging tests are only for smartcam hubs as smart hubs do not test categories
    if not isinstance(dev, SmartCamDevice):
        return

    # setup
    mock_child = copy.deepcopy(first_child)
    mock_components = copy.deepcopy(first_child_components)

    mock_child_device_info["child_device_list"] = [mock_child]
    mock_child_device_components["child_component_list"] = [mock_components]
    mock_child_device_info["sum"] = 1
    mock_child_device_components["sum"] = 1

    # Test can't find matching components

    mock_child["device_id"] = "no_comps_1"
    mock_components["device_id"] = "no_comps_2"

    caplog.set_level("DEBUG")
    caplog.clear()
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Could not find child components for device" in caplog.text

    caplog.clear()

    # Test doesn't log multiple
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Could not find child components for device" not in caplog.text

    # Test invalid category

    mock_child["device_id"] = "invalid_cat"
    mock_components["device_id"] = "invalid_cat"
    mock_child["category"] = "foobar"

    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Child device type not supported" in caplog.text

    caplog.clear()

    # Test doesn't log multiple
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Child device type not supported" not in caplog.text

    # Test no category

    mock_child["device_id"] = "no_cat"
    mock_components["device_id"] = "no_cat"
    mock_child.pop("category")

    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Child device type not supported" in caplog.text

    # Test only log once

    caplog.clear()
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Child device type not supported" not in caplog.text

    # Test no device_id

    mock_child.pop("device_id")

    caplog.clear()
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Could not find child id for device" in caplog.text

    # Test only log once

    caplog.clear()
    with patch.object(
        transport, "get_child_device_queries", side_effect=mock_get_child_device_queries
    ):
        await dev.update()

    assert "Could not find child id for device" not in caplog.text


@hubs_smart
async def test_unpair(dev: SmartDevice, mocker: MockerFixture):
    """Verify that unpair calls childsetup module."""
    if not dev.children:
        pytest.skip("device has no children")

    child = dev.children[0]

    assert child.parent is not None
    assert Module.ChildSetup in dev.modules
    cs = dev.modules[Module.ChildSetup]

    unpair_call = mocker.spy(cs, "unpair")

    unpair_feat = child.features.get("unpair")
    assert unpair_feat
    await unpair_feat.set_value(None)

    unpair_call.assert_called_with(child.device_id)
